/**
 * Copyright (c) 2005, Paul Tuckey
 * All rights reserved.
 * ====================================================================
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * ====================================================================
 */
package org.tuckey.web.filters.validation;

import org.tuckey.web.filters.validation.utils.BooleanUtils;
import org.tuckey.web.filters.validation.utils.Log;
import org.tuckey.web.filters.validation.utils.StringUtils;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletOutputStream;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;


/**
 * todo: add ant task that will validate a dir of files
 * todo: add sample xhtml file linked to from error page so user can copy an paste into theirs
 * todo: status code when failed parse should be 500 or similar
 */
public final class ValidationFilter implements Filter {

    public static final String VERSION = "0.6";

    private static Log log = Log.getLog(ValidationFilter.class);

    private Boolean enabled = null;
    private static final String VF_PROCESSED = "VF_PROCESSED";
    private static final String DEFAULT_ENABLE_ON_HOSTS = "localhost, local, 127.0.0.1";
    private boolean checkForDtdDefinition = true;
    private String xmlParser;
    private ServerNameMatcher serverNameMatcher;
    private String[] excludes;

    private boolean displayOriginal = true;

    /**
     * Gets run when the web application container initialises this filter.
     *
     * @param filterConfig the filter's config
     */
    public void init(FilterConfig filterConfig) throws ServletException {
        Log.setConfiguration(filterConfig);

        // Gets parameter from the server level config file
        String serverEnabledSetting = filterConfig.getServletContext().getInitParameter("enableValidationFilter");
        if (serverEnabledSetting != null) {
            this.enabled = (BooleanUtils.toBoolean(serverEnabledSetting)) ? Boolean.TRUE : Boolean.FALSE;
            log.info("ValidationFilter is " + (this.enabled.booleanValue() ? "" : "NOT ") + " enabled (by a server setting)");
        } else {
            // Gets parameter from the context web.xml file
            String contextEnabledSetting = filterConfig.getInitParameter("enabled");
            if (contextEnabledSetting != null) {
                this.enabled = (BooleanUtils.toBoolean(contextEnabledSetting)) ? Boolean.TRUE : Boolean.FALSE;
                log.info("ValidationFilter is " + (this.enabled.booleanValue() ? "" : "NOT ") + " enabled (by a context setting)");
            }
        }

        String checkForDtdDefinitionStr = filterConfig.getInitParameter("checkForDtdDefinition");
        if (checkForDtdDefinitionStr != null) {
            this.checkForDtdDefinition = BooleanUtils.toBoolean(checkForDtdDefinitionStr);
            log.info("checkForDtdDefinition is " + (this.checkForDtdDefinition ? "" : "NOT ") + " enabled");
        }

        String xmlParser = StringUtils.trimToNull(filterConfig.getInitParameter("xmlParser"));
        if (xmlParser != null) {
            this.xmlParser = xmlParser.trim();
            log.info("xmlParser is " + this.xmlParser);
            try {
                ValidationHandler.getCustomParserInstance(this.xmlParser);
            } catch (BadXmlParserException e) {
                throw new ServletException(e);
            }
        }

        String excludes = StringUtils.trimToNull(filterConfig.getInitParameter("excludes"));
        if (excludes != null) {
            this.excludes = excludes.split(",");
            for (int i = 0; i < this.excludes.length; i++) {
                String exclude = this.excludes[i];
                if ( exclude != null ) {
                    this.excludes[i] = exclude.trim();
                }
            }
            log.info("excludes are " + this.excludes);
        }

        String enableOnHosts = filterConfig.getInitParameter("enableOnHosts");
        if (StringUtils.isBlank(enableOnHosts)) enableOnHosts = DEFAULT_ENABLE_ON_HOSTS;
        serverNameMatcher = new ServerNameMatcher(enableOnHosts);

        String displayOriginalString = filterConfig.getInitParameter("displayOriginal");
        if ( BooleanUtils.toBoolean(displayOriginalString) ) {
            displayOriginal = true;
        }
    }

    /**
     * Gets run when the web application container destroys the filter.
     */
    public void destroy() {
        this.serverNameMatcher = null;
        this.enabled = null;
        this.xmlParser = null;
        this.displayOriginal = false;
    }

    /**
     * Gets run for each request mapped by this filter.
     *
     * @param servletRequest the request
     * @param servletResponse the response
     * @param filterChain the chain
     * @throws IOException
     * @throws ServletException
     */
    public void doFilter(ServletRequest servletRequest, ServletResponse servletResponse, FilterChain filterChain)
            throws IOException, ServletException {

        HttpServletRequest httpServletRequest = (HttpServletRequest) servletRequest;

        boolean showSourceOverride = "true".equals(servletRequest.getParameter("vf-src"));

        // check we are enabled and not already called in this filter chain
        if (isEnabled(httpServletRequest) && servletRequest.getAttribute(VF_PROCESSED) == null) {
            servletRequest.setAttribute(VF_PROCESSED, "true");

            HttpServletResponse httpServletResponse = (HttpServletResponse) servletResponse;

            ResponseWrapper responseWrapper = new ResponseWrapper(httpServletResponse, checkForDtdDefinition);
            filterChain.doFilter(httpServletRequest, responseWrapper);

            if (responseWrapper.isCommitted()) {
                log.debug("response is comitted, must be redirect or similar");
                return;
            }

            if (responseWrapper.isContentXHTML()) {

                String responseString = responseWrapper.outputStreamToString();

                if (responseWrapper.getStatus() != HttpServletResponse.SC_OK ) {
                    log.debug("response is only valid when status is OK, status is " + responseWrapper.getStatus());
                    writeToOut(httpServletResponse, responseString);
                    return;
                }

                String url = httpServletRequest.getRequestURI();

                ValidationHandler validationHandler;
                if (this.xmlParser == null) {
                    validationHandler = new ValidationHandler(responseString, url);

                } else {
                    try {
                        validationHandler = new ValidationHandler(responseString, url, this.xmlParser);
                    } catch (BadXmlParserException e) {
                        throw new ServletException(e);
                    }
                }

                validationHandler.parse();

                if (!validationHandler.isValid() || showSourceOverride) {
                    log.debug("doc is NOT valid show output");
                    if (httpServletRequest.getQueryString() != null)
                        url += "?" + httpServletRequest.getQueryString();

                    // make sure we don't get cached
                    httpServletResponse.setHeader("Cache-Control", "no-cache"); //HTTP 1.1
                    httpServletResponse.setHeader("Cache-Control", "post-check=0, pre-check=0");
                    httpServletResponse.setHeader("Expires", "0"); //prevents caching at the proxy server

                    if ( !displayOriginal ) {
                        String reportOutput = validationHandler.getReportDiv();
                        writeToOut(httpServletResponse, reportOutput);
                    }   else {
                        String reportOutputDiv = validationHandler.getReportDiv();
                        httpServletResponse.setContentLength(reportOutputDiv.length() + responseString.length());
                        ServletOutputStream sos = httpServletResponse.getOutputStream();
                        sos.write(responseString.getBytes());
                        sos.write(reportOutputDiv.getBytes());
                        sos.flush();
                    }

                } else {
                    log.debug("doc is valid passthrough");
                    writeToOut(httpServletResponse, responseString);
                }

            }   else {
                //log.debug("doc is not xhtml passthrough");
                //writeToOutRaw(httpServletResponse, responseCollector, servletOutputStream);
            }
        } else {
            filterChain.doFilter(servletRequest, servletResponse);
        }
    }

    private static void writeToOut(HttpServletResponse httpServletResponse, String responseString)
            throws IOException {
        httpServletResponse.setContentLength(responseString.length());
        ServletOutputStream sos = httpServletResponse.getOutputStream();
        sos.write(responseString.getBytes());
        sos.flush();
    }

    private boolean isEnabled(HttpServletRequest servletRequest) {
        // check to see if we are specifically enabled or not in web.xml or server level context conf
        if (enabled != null) {
            log.debug("enabled explicitly set to " + enabled.booleanValue());
            if ( ! enabled.booleanValue() ) {
                return false;
            }
        }

        if ( ! serverNameMatcher.isMatch(servletRequest.getServerName()) ) {
            log.debug("server name " + servletRequest.getServerName() + " does not match");
            return false;
        }

        // check the excludes
        if (excludes != null) {
            String url = servletRequest.getRequestURI();
            for (int i = 0; i < excludes.length; i++) {
                String exclude = excludes[i];
                if (url.matches(exclude)) {
                    if ( log.isDebugEnabled() ) {
                        log.debug("url" + url + " excluded because of " + exclude);
                    }
                    return false;
                }
            }
        }

        return true;
    }

}


